@@ -21,7 +21,6 @@
 #include "llvm/ADT/ilist.h"
 #include "llvm/CodeGen/MachineBasicBlock.h"
 #include "llvm/IR/DebugLoc.h"
-#include "llvm/IR/Metadata.h"
 #include "llvm/Support/Allocator.h"
 #include "llvm/Support/ArrayRecycler.h"
 #include "llvm/Support/Recycler.h"
@@ -38,12 +37,9 @@ class MachineJumpTableInfo;
 class MachineModuleInfo;
 class MCContext;
 class Pass;
-class PseudoSourceValueManager;
 class TargetMachine;
-class TargetSubtargetInfo;
 class TargetRegisterClass;
 struct MachinePointerInfo;
-struct WinEHFuncInfo;
 
 template <>
 struct ilist_traits<MachineBasicBlock>
@@ -74,24 +70,15 @@ private:
 /// MachineFunction is destroyed.
 struct MachineFunctionInfo {
   virtual ~MachineFunctionInfo();
-
-  /// \brief Factory function: default behavior is to call new using the
-  /// supplied allocator.
-  ///
-  /// This function can be overridden in a derive class.
-  template<typename Ty>
-  static Ty *create(BumpPtrAllocator &Allocator, MachineFunction &MF) {
-    return new (Allocator.Allocate<Ty>()) Ty(MF);
-  }
 };
 
 class MachineFunction {
   const Function *Fn;
   const TargetMachine &Target;
-  const TargetSubtargetInfo *STI;
   MCContext &Ctx;
   MachineModuleInfo &MMI;
-
+  GCModuleInfo *GMI;
+  
   // RegInfo - Information about each register in use in the function.
   MachineRegisterInfo *RegInfo;
 
@@ -104,14 +91,10 @@ class MachineFunction {
 
   // Keep track of constants which are spilled to memory
   MachineConstantPool *ConstantPool;
-
+  
   // Keep track of jump tables for switch instructions
   MachineJumpTableInfo *JumpTableInfo;
 
-  // Keeps track of Windows exception handling related data. This will be null
-  // for functions that aren't using a funclet-based EH personality.
-  WinEHFuncInfo *WinEHInfo = nullptr;
-
   // Function-level unique numbering for MachineBasicBlocks.  When a
   // MachineBasicBlock is inserted into a MachineFunction is it automatically
   // numbered and this vector keeps track of the mapping from ID's to MBB's.
@@ -137,7 +120,7 @@ class MachineFunction {
   /// this translation unit.
   ///
   unsigned FunctionNumber;
-
+  
   /// Alignment - The alignment of the function.
   unsigned Alignment;
 
@@ -151,24 +134,18 @@ class MachineFunction {
   /// True if the function includes any inline assembly.
   bool HasInlineAsm;
 
-  // Allocation management for pseudo source values.
-  std::unique_ptr<PseudoSourceValueManager> PSVManager;
-
-  MachineFunction(const MachineFunction &) = delete;
-  void operator=(const MachineFunction&) = delete;
+  MachineFunction(const MachineFunction &) LLVM_DELETED_FUNCTION;
+  void operator=(const MachineFunction&) LLVM_DELETED_FUNCTION;
 public:
   MachineFunction(const Function *Fn, const TargetMachine &TM,
-                  unsigned FunctionNum, MachineModuleInfo &MMI);
+                  unsigned FunctionNum, MachineModuleInfo &MMI,
+                  GCModuleInfo* GMI);
   ~MachineFunction();
 
   MachineModuleInfo &getMMI() const { return MMI; }
+  GCModuleInfo *getGMI() const { return GMI; }
   MCContext &getContext() const { return Ctx; }
 
-  PseudoSourceValueManager &getPSVManager() const { return *PSVManager; }
-
-  /// Return the DataLayout attached to the Module associated to this MF.
-  const DataLayout &getDataLayout() const;
-
   /// getFunction - Return the LLVM function that this machine code represents
   ///
   const Function *getFunction() const { return Fn; }
@@ -185,18 +162,6 @@ public:
   ///
   const TargetMachine &getTarget() const { return Target; }
 
-  /// getSubtarget - Return the subtarget for which this machine code is being
-  /// compiled.
-  const TargetSubtargetInfo &getSubtarget() const { return *STI; }
-  void setSubtarget(const TargetSubtargetInfo *ST) { STI = ST; }
-
-  /// getSubtarget - This method returns a pointer to the specified type of
-  /// TargetSubtargetInfo.  In debug builds, it verifies that the object being
-  /// returned is of the correct type.
-  template<typename STC> const STC &getSubtarget() const {
-    return *static_cast<const STC *>(STI);
-  }
-
   /// getRegInfo - Return information about the registers currently in use.
   ///
   MachineRegisterInfo &getRegInfo() { return *RegInfo; }
@@ -209,7 +174,7 @@ public:
   MachineFrameInfo *getFrameInfo() { return FrameInfo; }
   const MachineFrameInfo *getFrameInfo() const { return FrameInfo; }
 
-  /// getJumpTableInfo - Return the jump table info object for the current
+  /// getJumpTableInfo - Return the jump table info object for the current 
   /// function.  This object contains information about jump tables in the
   /// current function.  If the current function has no jump tables, this will
   /// return null.
@@ -220,18 +185,13 @@ public:
   /// does already exist, allocate one.
   MachineJumpTableInfo *getOrCreateJumpTableInfo(unsigned JTEntryKind);
 
+  
   /// getConstantPool - Return the constant pool object for the current
   /// function.
   ///
   MachineConstantPool *getConstantPool() { return ConstantPool; }
   const MachineConstantPool *getConstantPool() const { return ConstantPool; }
 
-  /// getWinEHFuncInfo - Return information about how the current function uses
-  /// Windows exception handling. Returns null for functions that don't use
-  /// funclets for exception handling.
-  const WinEHFuncInfo *getWinEHFuncInfo() const { return WinEHInfo; }
-  WinEHFuncInfo *getWinEHFuncInfo() { return WinEHInfo; }
-
   /// getAlignment - Return the alignment (log2, not bytes) of the function.
   ///
   unsigned getAlignment() const { return Alignment; }
@@ -274,7 +234,7 @@ public:
   template<typename Ty>
   Ty *getInfo() {
     if (!MFInfo)
-      MFInfo = Ty::template create<Ty>(Allocator, *this);
+      MFInfo = new (Allocator.Allocate<Ty>()) Ty(*this);
     return static_cast<Ty*>(MFInfo);
   }
 
@@ -295,19 +255,19 @@ public:
   }
 
   /// Should we be emitting segmented stack stuff for the function
-  bool shouldSplitStack() const;
+  bool shouldSplitStack();
 
   /// getNumBlockIDs - Return the number of MBB ID's allocated.
   ///
   unsigned getNumBlockIDs() const { return (unsigned)MBBNumbering.size(); }
-
+  
   /// RenumberBlocks - This discards all of the MachineBasicBlock numbers and
   /// recomputes them.  This guarantees that the MBB numbers are sequential,
   /// dense, and match the ordering of the blocks within the function.  If a
   /// specific MachineBasicBlock is specified, only that block and those after
   /// it are renumbered.
   void RenumberBlocks(MachineBasicBlock *MBBFrom = nullptr);
-
+  
   /// print - Print out the MachineFunction in a format suitable for debugging
   /// to the specified stream.
   ///
@@ -342,12 +302,6 @@ public:
   typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
   typedef std::reverse_iterator<iterator>             reverse_iterator;
 
-  /// Support for MachineBasicBlock::getNextNode().
-  static BasicBlockListType MachineFunction::*
-  getSublistAccess(MachineBasicBlock *) {
-    return &MachineFunction::BasicBlocks;
-  }
-
   /// addLiveIn - Add the specified physical register as a live-in value and
   /// create a corresponding virtual register for it.
   unsigned addLiveIn(unsigned PReg, const TargetRegisterClass *RC);
@@ -380,21 +334,15 @@ public:
   void splice(iterator InsertPt, iterator MBBI) {
     BasicBlocks.splice(InsertPt, BasicBlocks, MBBI);
   }
-  void splice(iterator InsertPt, MachineBasicBlock *MBB) {
-    BasicBlocks.splice(InsertPt, BasicBlocks, MBB);
-  }
   void splice(iterator InsertPt, iterator MBBI, iterator MBBE) {
     BasicBlocks.splice(InsertPt, BasicBlocks, MBBI, MBBE);
   }
 
-  void remove(iterator MBBI) { BasicBlocks.remove(MBBI); }
-  void remove(MachineBasicBlock *MBBI) { BasicBlocks.remove(MBBI); }
-  void erase(iterator MBBI) { BasicBlocks.erase(MBBI); }
-  void erase(MachineBasicBlock *MBBI) { BasicBlocks.erase(MBBI); }
-
-  template <typename Comp>
-  void sort(Comp comp) {
-    BasicBlocks.sort(comp);
+  void remove(iterator MBBI) {
+    BasicBlocks.remove(MBBI);
+  }
+  void erase(iterator MBBI) {
+    BasicBlocks.erase(MBBI);
   }
 
   //===--------------------------------------------------------------------===//
@@ -451,9 +399,9 @@ public:
   MachineMemOperand *getMachineMemOperand(MachinePointerInfo PtrInfo,
                                           unsigned f, uint64_t s,
                                           unsigned base_alignment,
-                                          const AAMDNodes &AAInfo = AAMDNodes(),
+                                          const MDNode *TBAAInfo = nullptr,
                                           const MDNode *Ranges = nullptr);
-
+  
   /// getMachineMemOperand - Allocate a new MachineMemOperand by copying
   /// an existing one, adjusting by an offset and using the given size.
   /// MachineMemOperands are owned by the MachineFunction and need not be
@@ -503,19 +451,16 @@ public:
     extractStoreMemRefs(MachineInstr::mmo_iterator Begin,
                         MachineInstr::mmo_iterator End);
 
-  /// Allocate a string and populate it with the given external symbol name.
-  const char *createExternalSymbolName(StringRef Name);
-
   //===--------------------------------------------------------------------===//
   // Label Manipulation.
   //
-
+  
   /// getJTISymbol - Return the MCSymbol for the specified non-empty jump table.
   /// If isLinkerPrivate is specified, an 'l' label is returned, otherwise a
   /// normal 'L' label is returned.
-  MCSymbol *getJTISymbol(unsigned JTI, MCContext &Ctx,
+  MCSymbol *getJTISymbol(unsigned JTI, MCContext &Ctx, 
                          bool isLinkerPrivate = false) const;
-
+  
   /// getPICBaseSymbol - Return a function-local symbol to represent the PIC
   /// base.
   MCSymbol *getPICBaseSymbol() const;
